## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# SIMULATOR.py --- 'Plain' DEVS Model Simulator
#                     --------------------------------
#                            Copyright (c) 2000
#                          Jean-Sebastien  BOLDUC
#                             Hans  Vangheluwe
#                       McGill University (Montreal)
#                     --------------------------------
# Version 1.0                                        last modified: 01/11/01
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
#
# GENERAL NOTES AND REMARKS:
#
# Atomic- and coupled-DEVS specifications adapted from:
#       B.P.Zeigler, ''Theory of modeling and simulation: Integrating
#       Discrete Event and Continuous Complex Dynamic Systems '',
#       Academic Press, 2000
#
# Description of message passing mechanism (to be completed)
#
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##



##  GLOBAL VARIABLES AND FUNCTIONS
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

# Necessary to send COPIES of messages.
from copy import deepcopy
from DevsSimInterfaces import DevsEvent
from threading import *
import time
from DEVS import *

#done = False
#lock = Lock()

# {\tt __VERBOSE__} switch allows to enter\dots well, ''verbose mode'' when set
# to true. Useful for debugging (similar to {\tt #define _DEBUG_} convention in
# {\tt C++}).
__VERBOSE__ = 0

def Error(message = '', esc = 1):
  """Error-handling function: reports an error and exits interpreter if
  {\tt esc} evaluates to true.
  
  To be replaced later by exception-handling mechanism.
  """
  from sys import exit, stderr
  stderr.write("ERROR: %s\n" % message)
  if esc:
    exit(1)

##  SIMULATOR CLASSES
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

class AtomicSolver(AtomicDEVS):
  """Simulator for atomic-DEVS.
  
  Atomic-DEVS can receive three types of messages: $(i,\,t)$, $(x,\,t)$ and
  $(*,\,t)$. The latter is the only one that triggers another message-
  sending, namely, $(y,\,t)$ is sent back to the parent coupled-DEVS. This is
  actually implemented as a return value (to be completed).
  """
  
  ###
  def receive(self, msg):
    
    # For any received message, the time {\tt t} (time at which the message
    # is sent) is the second item in the list {\tt msg}.
    t = msg[1]
  
    # $(i,\,t)$ message --- sets origin of time at {\tt t}:
    if msg[0] == 0:
      #print 'msg==0'
      self.timeLast = t - self.elapsed 
      self.timeNext = self.timeLast + self.timeAdvance()
      #print self.timeNext
      evt = DevsEvent(self.state, 'Initialization', t)
      self.newEvent(evt)
      print self.instName, self.handler.listeners
  
    # $(*,\,t)$ message --- triggers internal transition and returns
    # $(y,\,t)$ message for parent coupled-DEVS:
    elif msg[0] == 1:
      #print 'msg == 1'
      if t != self.timeNext:
        Error("Bad synchronization...", 1)

      # First call the output function, which (amongst other things) rebuilds
      # the output dictionnary {\tt myOutput}:
      self.myOutput = {}
      self.outputFnc()
                
      # Call the internal transition function to update the DEVS' state, and
      # update time variables:
      self.state = self.intTransition()      
      self.timeLast = t
      self.timeNext = self.timeLast + self.timeAdvance()      
      self.elapsed = 0.
      
      if __VERBOSE__:
        print "\n\tINTERNAL TRANSITION: %s" % (self.myID)
        print "\t  New State: %s" % (self.state)
        print "\t  Output Port Configuration:"
        for I in range(0, len(self.OPorts) ):
          if self.OPorts[I] in self.myOutput.keys():
            print "\t    port%d: %s at %s" % (I, self.myOutput[self.OPorts[I]], \
              repr(self.myOutput[self.OPorts[I]]))
          else:
            print "\t    port%d: None" % (I)
        print "\t  Next scheduled internal transition at time %f" % self.timeNext
      
      # Return the DEVS' output to the parent coupled-DEVS (rather than
      # sending $(y,\,t)$ message).
      
#      evt = DevsEvent(self.instType+'.'+self.instName, 'Internal\nTransition', t)
      evt = DevsEvent(self.state, 'Internal\nTransition', t)
      self.newEvent(evt)      
      print evt

      return self.myOutput

    # ${x,\,t)$ message --- triggers external transition, where $x$ is the
    # input dictionnary to the DEVS:
    elif type(msg[0]) == type({}):
      if not(self.timeLast <= t <= self.timeNext):
        Error("Bad synchronization...", 1)
      
      self.myInput = msg[0]
      # update elapsed time. This is necessary for the call to the external
      # transition function, which is used to update the DEVS' state.
      self.elapsed = t - self.timeLast
      self.state = self.extTransition()      
      # Udpate time variables:
      self.timeLast = t
      self.timeNext = self.timeLast + self.timeAdvance()      
      self.elapsed = 0
      
      if __VERBOSE__:
        print "\n\tEXTERNAL TRANSITION: %s" % (self.myID)
        print "\t  Input Port Configuration:"
        for I in range(0, len(self.IPorts) ):
          print "\t    port%d: %s at %s" % (I, self.peek(self.IPorts[I]),\
            repr(self.peek(self.IPorts[I])))
        print "\t  New State: %s" % (self.state)
        print "\t  Next scheduled internal transition at time %f" % self.timeNext

      evt = DevsEvent(self.state, 'External\nTransition', t)
      self.newEvent(evt)
            
    else:
      Error("Unrecognized message", 1)

#    ===================================================================    #

class CoupledSolver(CoupledDEVS):
  """Simulator (coordinator) for coupled-DEVS.
  
  Coupled-DEVS can receive the same three types of messages as for atomic-
  DEVS, plus the $(y,\,t)$ message. The latter is implemented as a returned
  value rather than a message. This is possible because the $(y,\,t)$ message
  is always sent to a coupled-DEVS in response from its sending a $(*,\,t)$
  message. (This implementation makes it possible to easily distinguish
  $(y,\,t)$ from $(x,\,t)$ messages... to be completed)
  Description of {\tt eventList} and {\tt dStar} (to be completed)
  """

  ###
  def receive(self, msg):
    
    # For any received message, the time {\tt t} (time at which the message 
    # is sent) is the second item in the list {\tt msg}.
    t = msg[1]
    
    # $(i,\,t)$ message --- sets origin of time at {\tt t}:
    if msg[0] == 0:
      
      # Rebuild event-list and update time variables, by sending the
      # initialization message to all the children of the coupled-DEVS. Note
      # that the event-list is not sorted here, but only when the list of
      # {\sl imminent children\/} is needed. Also note that {\tt None} is
      # defined as bigger than any number in Python (stands for $+\infty$).
      self.timeLast = 0.
      self.timeNext = 100000
      self.eventList = []
      for d in self.componentSet:
        self.send(d, msg)
        self.eventList.append( [d.timeNext, d] )
        self.timeNext = min(self.timeNext, d.timeNext)
        self.timeLast = max(self.timeLast, d.timeLast)
    # $(*,\,t)$ message --- triggers internal transition and returns
    # $(y,\,t)$ message for parent coupled-DEVS:
    elif msg[0] == 1:
      if t != self.timeNext:
        Error("Bad synchronization...", 1)
        
      # Build the list {\tt immChildren} of {\sl imminent children\/} based
      # on the sorted event-list, and select the active-child {\tt dStar}. If
      # there are more than one imminent child, the coupled-DEVS {\tt select}
      # function is used to decide the active-child.
      self.eventList.sort()
      immChildren = []
      for I in self.eventList:
        if I[0] == t:
          immChildren.append(I[1])
        else:
          break
      if len(immChildren) == 1:
        dStar = immChildren[0]
      else:
        dStar = self.select(immChildren)
        if __VERBOSE__:
          print "\n\tCollision occured in %s, involving:" % (self.myID)
          for I in immChildren:
            print "    \t   %s" % (I.myID)
          print "\t  select chooses %s" % (dStar.myID)

      inst = dStar.instType+'.'+dStar.instName
      evt = DevsEvent(inst, inst + '\'s\nTransition', t)
      self.newEvent(evt)      
        
      # Send $(*,\,t)$ message to active children, which returns (or sends
      # back) message $(y,\,t)$. In the present implementation, just the
      # sub-DEVS output dictionnary $y$ is returned and stored in {\tt Y}:
      Y = self.send(dStar, msg)
      
      # Scan the list of sub-DEVS to find those influenced by (connected to)
      # the active child. To these sub-DEVS a message $(x,\,t)$ is sent. The
      # generation of the input dictionnary {\tt X} relies on ports'
      # {\tt inLine} attributes rather than on coupled-DEVS' {\tt IC}. The
      # coupled-DEVS time variables are also updated in parallel. This piece
      # of code, although functionnal, could be far more elegant 
      # and efficient\dots
      self.timeLast = t
      self.timeNext = 100000
      self.eventList = []
      for d in self.componentSet:
        X = {}
        for p in d.IPorts:
          for pi in p.inLine:
            if pi in Y.keys():
              X[p] = deepcopy(Y[pi]) # only if atomic DEVS?
              break
#          if p.inLine in Y.keys():
#            X[p] = Y[p.inLine]
        if X != {}:
          self.send(d, [X, t])
        self.eventList.append( [d.timeNext, d] )
        self.timeNext = min(self.timeNext, d.timeNext)
        
      # Use same pattern as above to update coupled-DEVS output dictionnary
      # (once again using ports' {\tt inLine} attributes rather than
      # coupled-DEVS' {\tt EOC}). The DEVS' output is then returned to its
      # parent coupled-DEVS (rather than being sent).
      self.myOutput = {}
      for p in self.OPorts:
        for pi in p.inLine:
         if pi in Y.keys():
           self.myOutput[p] = Y[pi]
           break
#        if p.inLine in Y.keys():
#          self.myOutput[p] = Y[p.inLine]

      return self.myOutput
    
    # ${x,\,t)$ message --- triggers external transition, where $x$ is the
    # input dictionnary to the DEVS:
    elif type(msg[0]) == type({}):
      if not(self.timeLast <= t <= self.timeNext):
        Error("Bad synchronization...", 1)
      
      self.myInput = msg[0]      

      # Send $(x,\,t)$ messages to those sub-DEVS influenced by the coupled-
      # DEVS input ports (ref. {\tt EIC}). This is essentially the same code
      # as above that also updates the coupled-DEVS' time variables in
      # parallel.
      self.timeLast = t
      self.timeNext = 100000
      self.eventList = []
      for d in self.componentSet:
        X = {}
        for p in d.IPorts:
          for pi in p.inLine:
            if pi in self.myInput.keys():
              X[p] = self.myInput[pi]
              break
#          if p.inLine in self.myInput.keys():
#            X[p] = self.myInput[p.inLine]
        if X != {}:
          self.send(d, [X, t])
        self.eventList.append( [d.timeNext, d] )
        self.timeNext = min(self.timeNext, d.timeNext)

    else:
      Error("Unrecognized message", 1)

  def send(self, d, msg):
    """Dispatch messages to the right method.
    """
    return d.receive(msg)
#    ===================================================================    #

class Simulator(Thread, AtomicSolver, CoupledSolver):
  """Associates a hierarchical DEVS model with the simulation engine.
  """
  
  ###
  def __init__(self, model, time=100, isRealTime=False):
    """Constructor.
  
    {\tt model} is an instance of a valid hierarchical DEVS model. The
    constructor stores a local reference to this model and augments it with
    {\sl time variables\/} required by the simulator.
    """
  
    self.model      = model
    self.simTime    = time
    self.isRealTime = isRealTime
    self.lock       = Lock()
    self.done       = False
    Thread.__init__(self)
    ###self.augment(self.model)  #Time variabels have been added to the models directly
                                 #Hongyan Song, August 2nd, 2005
    #if(self.isRealTime): import time  
  ###
  def augment(self, d):
    """Recusively augment model {\tt d} with {\sl time variables\/}.
    """
   
    # {\tt timeLast} holds the simulation time when the last event occured
    # within the given DEVS, and (\tt timeNext} holds the scheduled time for
    # the next event.
    d.timeLast = d.timeNext = 0. 
    if d.type() == 'COUPLED':
      # {\tt eventList} is the list of pairs $(tn_d,\,d)$, where $d$ is a
      # reference to a sub-model of the coupled-DEVS and $tn_d$ is $d$'s
      # time of next event.
      d.eventList = []
      for subd in d.componentSet:
        self.augment(subd)
        
  def setTime(self, time):
    self.simTime = time
  
  def setDone(self):
    self.lock.acquire()
    self.done = True
    self.lock.release()

  def send(self, d, msg):
    """Dispatch messages to the right method.
    """
    return d.receive(msg)
#    if d.type() == 'COUPLED':
#      print 'Coupled', d.instName
#      return d.receive(msg)
#    elif d.type() == 'ATOMIC':
#      print 'Atomic', d.instName
#      evt = DevsEvent(d.state, 'Atomic\nTransition', t)
#      d.newEvent(evt)
#      return d.receive(msg)

  ###
  def run(self):
    """Simulate the model (Root-Coordinator).
    
    In this implementation, the simulation stops only when the simulation
    time (stored in {\tt clock}) reaches {\tt T} (starts at 0). Other means
    to stop the simulation ({\it e.g.}, when an atomic-DEVS enters a given
    state, or when a global variable reaches a predefined value) will be
    implemented.
    """
  
    # Initialize the model --- set the simulation clock to 0.
    self.send(self.model, (0, 0))

    if(self.isRealTime):
      time.sleep(self.model.timeNext)
    clock = self.model.timeNext

    # Main loop repeatedly sends $(*,\,t)$ messages to the model's root
    # DEVS.
    while((clock <= self.simTime) and (not self.done)):
      #print 'done:', self.done
      print "\n", "* " * 10, "CLOCK: %f" % clock, "\n"
      self.send(self.model, (1, clock))

      if __VERBOSE__:
        print "\n\tROOT DEVS' OUTPUT PORT CONFIGURATION:"
        for I in range(0, len(self.model.OPorts) ):
          if self.model.OPorts[I] in self.model.myOutput.keys():
            print "\t    port%d: %s at %s" % (I, \
              self.model.myOutput[self.model.OPorts[I]], \
              repr(self.model.myOutput[self.model.OPorts[I]]))
          else:
            print "\t    port%d: None" % (I)

      if(self.isRealTime):            
        time.sleep(self.model.timeNext-clock)       
      clock = self.model.timeNext

