# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

from AST.ASTExp import *
from FrontEnd.printers.EqExpPrinter import EqExpPrinter
from BackEnd.causality.Node import Node, EqNode, VarNode
from BackEnd.causality.Graph import FlowGraph
from BackEnd.causality.Edge import FlowEdge
from BackEnd.Exception.CausalityExceptions import InconsistantModelException

class CausalityAssigner:
    mainModel=None # RegularClassDefinition

    def __init__(self, ast):
        self.mainModel=ast.classFiles[0].classDefinitions[0]
	self.msg=""

    # Generate all r-element combinations of the elements of the list set
    # This code is from:
    # http://www2.bc.edu/~fisherbb/Classes/Fall02/MC248/notes.html
    def comb(self, r, set):
        n = len(set)
        list = []
        if r < 0 or r > n:
            return list
        if r == 0:
            return [[]]
        x = set.pop(0)          # Remove the 0'th element.
        sublist = self.comb(r-1, set)
        for p in sublist:
            p.insert(0, x)  # p = [x] + p
        list.extend(sublist)    # list = list + sublist
        sublist = self.comb(r, set)
        list.extend(sublist)    # list = list + sublist
        set.insert(0, x)        # Put it back.
        return list

    def go(self):
	G=self.buildGraph(self.mainModel)
	#print '---------------dump original graph----------------'
        #print self.dumpGraph(G)
	G=self.assignCausality(G)
	#print '\n-------causality assignment result -------'
	self.msg=self.msg+"\n-------causality assignment result -------\n"
	self.msg=self.msg+self.dumpCausalityResult(G)+"\n"
	if not self.isValidCausality(G):
	    #print '***Invalid causaliy!***'
	    self.msg=self.msg+'***Invalid causaliy!***\n'
	    derNodes=G.getVarNodesWithDerForm()
	    numOfDerNodes=len(derNodes)
	    for numOfDerCausality in range(1, numOfDerNodes+1):
		combination=self.comb(numOfDerCausality, derNodes)
		for i in combination:
		    # reset the graph
		    G.reset() 
		    for j in i: # j: VarNode
			j.integralCausality=False
		    self.resetGraphEdges(G)
		    G=self.assignCausality(G)
		    #print '\n-------causality assignment result -------'
		    self.msg=self.msg+\
			     "\n-------causality assignment result -------\n"
		    self.msg=self.msg+self.dumpCausalityResult(G)+"\n"
		    if self.isValidCausality(G):
			break
		    else:
			#print '***Invalid causaliy!***'
			self.msg=self.msg+'***Invalid causaliy!***\n'
		if self.isValidCausality(G):
		    break
	return G

    # assigns causality
    def assignCausality(self, G):
	L=self.buildLevelGraph(G)
	while L.isSinkInL:
	    G=self.findBlockingFlow(L)
	    L=self.buildLevelGraph(G)
	return G

    # create a bipartite graph
    def buildGraph(self, classDef):
	G=FlowGraph()
	# source and sink use the reserved ids '0' & '-1'
	G.source=Node('0')
	G.sink=Node('-1')
	# create and add VarNodes to G
	for i in self.mainModel.getFlattenedComponents():
	    # parameters and constants are not carried into 
	    # causality assignment
	    if not (i.isParameter() or i.isConstant()):
            	newVarNode=VarNode(str(i.getName()), i)
            	G.addVarNode(newVarNode.getId(), newVarNode)
	# search through eqs for derivative form of vars
	for i in self.mainModel.getFlattenedEqs():
	    self.setDerivativeForm(G, i.getLHS())
        # how to handle initial eqs?
        id=1
        for i in self.mainModel.getFlattenedEqs():
            newEqNode=EqNode(str(id), i)
	    G.addEqNode(str(id), newEqNode)
            self.addOutgoingEdgesToEqNode(G, newEqNode, i.getLHS())
            id=id+1
        # link the source node and EqNodes
        source=G.getSource()
        sourceId=source.getId()
        for i in G.getEqNodes():
            source.addOutgoingEdgeInG(i.getId(), FlowEdge(i))
	    i.addOutgoingEdgeInG(source.getId(), \
				FlowEdge(source, capP=0, flowP=0))
	# delete isolated VarNodes from graph
        # link the set of VarNodes and the sink
        sink=G.getSink()
        sinkId=sink.getId()
	varNodeList=G.getVarNodes()
	numOfVars=len(varNodeList)
	index=numOfVars-1
	while index>=0:
	    if varNodeList[index].isIsolated:
		G.deleteVarNode(varNodeList[index].getId())
	    else:
            	varNodeList[index].addOutgoingEdgeInG(sinkId, FlowEdge(sink))
	    	sink.addOutgoingEdgeInG(varNodeList[index].getId(), \
			FlowEdge(varNodeList[index], capP=0, flowP=0))
	    index=index-1
	# validate that the number of EqNodes equals to the # of VarNodes
	numVars=len(G.getVarNodes())
	numEqs=len(G.getEqNodes())
	if numEqs!=numVars:
	    #raise 'Error: number of equations and variables are not equal!'
	    raise InconsistantModelException(numVars, numEqs)
	return G

    # rebuild adjacencies between EqNodes and VarNodes 
    # based on different combinations of derivative causality
    def resetGraphEdges(self, G):
	for node in G.getEqNodes():
	    node.deleteAllOutgoingEdgesInG()
	for node in G.getEqNodes():
	    self.addOutgoingEdgesToEqNode(G, node, node.getEq().getLHS())    
	# link the source node and EqNodes
        source=G.getSource()
        sourceId=source.getId()
        for i in G.getEqNodes():
            i.addOutgoingEdgeInG(source.getId(), \
                                FlowEdge(source, capP=0, flowP=0))

    # search through the LHS of an equation to mark variables which have 
    # derivative form
    def setDerivativeForm(self, G, exp):
	if isinstance(exp, IdentExp):
	    return 
        elif isinstance(exp, CompRefExp):
	    return
        elif isinstance(exp, FunCallExp):
	    operands=exp.getOperands()
	    if str(operands[0])=='der':
		args=operands[1].getOperands()
		if len(args)!=1:
		    raise 'der operator only has 1 argument!'
		else:
		    if isinstance(args[0], IdentExp) or \
		       isinstance(args[0], CompRefExp):
		    	varNode=G.varNodeLookup(str(args[0]))
			if varNode!=None:
			    varNode.hasDerivativeForm=True
			    # all nodes are initialized with integral causaity
			    varNode.integralCausality=True
        elif isinstance(exp, OrExp) or \
           isinstance(exp, AndExp) or \
           isinstance(exp, NotExp) or \
           isinstance(exp, LessExp) or \
           isinstance(exp, LessEQExp) or \
           isinstance(exp, GreaterExp) or \
           isinstance(exp, GreaterEQExp)or \
           isinstance(exp, EQEQExp) or \
           isinstance(exp, LessGtExp):
            for i in exp.getOperands():
                self.setDerivativeForm(G, i)
	elif isinstance(exp, SubtractExp) or \
             isinstance(exp, SumExp) or \
             isinstance(exp, MulExp) or \
             isinstance(exp, DivExp) or \
             isinstance(exp, UnaryMinusExp) or \
             isinstance(exp, PowerExp) or \
             isinstance(exp, RangeExp):
            for i in exp.getOperands():
                self.setDerivativeForm(G, i)
        elif  isinstance(exp, FunArgsExp) or \
             isinstance(exp, NamedArgExp) or \
             isinstance(exp, ExpListExp):
            for i in exp.getOperands():
                self.setDerivativeForm(G, i)
        else:
            return

    # add edges from EqNodes to VarNodes
    def addOutgoingEdgesToEqNode(self, G, eqNode, exp):
        if isinstance(exp, IdentExp):
            # the IdentExp must be a declared variable
            # eg., "der" is an IdentExp, but not variable
	    varNode=G.varNodeLookup(str(exp))
            if varNode!=None and varNode.integralCausality==False:
                eqNode.addOutgoingEdgeInG(varNode.getId(), FlowEdge(varNode))
		reversedEdge=FlowEdge(eqNode, capP=0, flowP=0)
		varNode.addOutgoingEdgeInG(eqNode.getId(), reversedEdge)
		varNode.setIsolated(False)
        elif isinstance(exp, CompRefExp):
            for i in exp.getOperands():
                self.addOutgoingEdgesToEqNode(G, eqNode, i)
	elif isinstance(exp, FunCallExp):
	    # this only handles the 'der' operator currently
	    operands=exp.getOperands()
            if str(operands[0])=='der':
                args=operands[1].getOperands()
                if len(args)!=1:
                    raise 'der operator only has 1 argument!'
                else:
                    if isinstance(args[0], IdentExp) or \
                       isinstance(args[0], CompRefExp):
                        varNode=G.varNodeLookup(str(args[0]))
                        if varNode!=None and varNode.integralCausality==True:
			    eqNode.addOutgoingEdgeInG(varNode.getId(), \
				FlowEdge(varNode))
			    reversedEdge=FlowEdge(eqNode, capP=0, flowP=0)
			    varNode.addOutgoingEdgeInG(eqNode.getId(), \
				reversedEdge)
			    varNode.setIsolated(False)
	    else:
		#raise 'need handle other kind of funtions, eg., sin() ...'
		funArgs=operands[1].getOperands()
		for i in funArgs:
		    self.addOutgoingEdgesToEqNode(G, eqNode, i)
        elif isinstance(exp, SingleCompRefExp):
            operands=exp.getOperands()
	    varNode=G.varNodeLookup(str(operands[0]))
	    if varNode!=None:
		eqNode.addOutgoingEdgeInG(varNode.getId(), FlowEdge(varNode))
		reversedEdge=FlowEdge(eqNode, capP=0, flowP=0)
		varNode.addOutgoingEdgeInG(eqNode.getId(), reversedEdge)
		varNode.setIsolated(False)
	elif isinstance(exp, OrExp) or \
           isinstance(exp, AndExp) or \
           isinstance(exp, NotExp) or \
           isinstance(exp, LessExp) or \
           isinstance(exp, LessEQExp) or \
           isinstance(exp, GreaterExp) or \
           isinstance(exp, GreaterEQExp)or \
           isinstance(exp, EQEQExp) or \
           isinstance(exp, LessGtExp):
            for i in exp.getOperands():
                self.addOutgoingEdgesToEqNode(G, eqNode, i)
        elif isinstance(exp, SubtractExp) or \
             isinstance(exp, SumExp) or \
             isinstance(exp, MulExp) or \
             isinstance(exp, DivExp) or \
             isinstance(exp, UnaryMinusExp) or \
             isinstance(exp, PowerExp) or \
             isinstance(exp, RangeExp):
            for i in exp.getOperands():
                self.addOutgoingEdgesToEqNode(G, eqNode, i)
        #elif isinstance(exp, IfElseExp):
        elif isinstance(exp, NamedArgExp):
	    # vars only appear on the LHS of an NameArgExp
	    self.addOutgoingEdgesToEqNode(G, eqNode, exp.getOperands()[0])
        elif isinstance(exp, ExpListExp):
            for i in exp.getOperands():
                self.addOutgoingEdgesToEqNode(G, eqNode, i)
	else:
            pass

    # create the level graph L using BFS
    def buildLevelGraph(self, G):
	for node in G.getAllNodes():
	    node.deleteAllOutgoingEdgesInL()
	    node.setVisited(False)
	    node.setLevel(-1000000)
	G.isSinkInL=False
	G.source.setVisited(True)
	L0=[G.source]
	level=0
	G.source.setLevel(level)
	while len(L0)>0:
	    level=level+1
	    L1=[]
	    for v1 in L0:
		for e in v1.getAllUnsaturatedEdgesInG():
		    w=e.getDestination()
		    if not w.isVisited():
			w.setVisited(True)
			w.setLevel(level)
			# set the flag if the sink is in L
			if w==G.sink:
			    G.isSinkInL=True
			L1.append(w)
			v1.addOutgoingEdgeInL(w.getId(), e)
		    else:
			if v1.level+1==w.level:
			    v1.addOutgoingEdgeInL(w.getId(), e)
	    L0=L1
	return G

    # find a blocking flow on the level graph L for the current flow
    def findBlockingFlow(self, L):
	# initialize
	p=[L.source]  # path
	v=L.source
	while True:
	    # Advance
	    outgoingEdges=v.getAllUnsaturatedEdgesInL()
	    if len(outgoingEdges)>0: 
		w=outgoingEdges[0].getDestination()
		p.append(w)
		v=w
		if v!=L.sink:
		    continue # repeat Advance
	    else: # Retreat
		if v==L.source:
		    return L # halt
		else:
		    v=p[len(p)-1]
		    u=p[len(p)-2]
		    u.deleteOutgoingEdgeInL(v.getId())
		    p=p[0:len(p)-1]
		    v=u
		    continue
	    # Augment
	    delta=self.calSaturatedFlow(p)
	    # add delta to the flow of every edge on path p
	    # and delete all newly saturated edge in L
	    index=0
	    while index<=len(p)-2:
		n1=p[index]
		n2=p[index+1]
		# add delta to the flow of every edge on path p in L
		e1=n1.getOutgoingEdgeInL(n2.getId())
		e1.addFlow(delta)
		# delta has also been added to the flow of 
		# every edge on p in G
		# add -delta to the flow of corresponding reversed edge
		e3=n2.getOutgoingEdgeInG(n1.getId())
		e3.addFlow(-delta)
		#delete all newly saturated edge in L
		if e1.getCap()==e1.getFlow():
		    n1.deleteOutgoingEdgeInL(n2.getId())
		index=index+1
	    # initialize
	    p=[L.source]
	    v=L.source
	return L

    # compute the minimum of (cap(v,w)-f(v,w)) for [v,w] an edge of p
    # p: [] of nodes
    def calSaturatedFlow(self, p):
	index=0
	min=1000000
	while index<=len(p)-2:
	    n1=p[index]
	    n2=p[index+1]
	    e=n1.getOutgoingEdgeInL(n2.getId())
	    res=e.getCap()-e.getFlow()
	    if res<min:
		min=res
	    index=index+1
	return min

    # validate the causality result
    def isValidCausality(self, G):
	for i in G.getEqNodes():
	    if len(i.getAllSaturatedEdgesInG())!=1:
		return False
	return True


    def dumpGraph(self, G):
        result='source: \n  '+str(G.source)+'\n'
	for i in G.source.getAllOutgoingEdgesInG():
	    result=result+'    '+str(i)+'\n'
        result=result+'eqNodes: \n'
        for i in G.getEqNodes():
	#    result=result+'  '+str(i)+'\n'
            result=result+'  '+str(i)+'\n'
            #result=result+'     Outgoing Edges: \n'
            for j in i.getAllOutgoingEdgesInG():
                result=result+'    '+str(j)+'\n'
        result=result+'VarNodes: \n'
        for i in G.getVarNodes():
	    #if i.integralCausality==True:
		#result=result+' der('+str(i)+')\n'
	    #else:
	    result=result+'  '+str(i)+'\n'
            #result=result+'     Outgoing Edges: '
            for j in i.getAllOutgoingEdgesInG():
                result=result+'    '+str(j)+'\n'
	result=result+'sink: \n  '+str(G.sink)+'\n'
	for i in G.sink.getAllOutgoingEdgesInG():
	    result=result+'    '+str(i)+'\n'
        return result

    def dumpLevelGraph(self, L):
	result='source: \n  '+L.source.dumpL()+'\n'
	result=result+'eqNodes: \n'
	for i in L.getEqNodes():
	    result=result+'  '+i.dumpL()+'\n'
	result=result+'VarNodes: \n'
	for i in L.getVarNodes():
	    result=result+'  '+i.dumpL()+'\n'
	result=result+'sink: \n  '+L.sink.dumpL()
	return result

    def dumpCausalityResult(self, G): #causality):
	printer=EqExpPrinter()
	vars='Variables: ' 
	for i in G.getVarNodes():
	    vars=vars+str(i)+', '
	#print vars
	msg=vars+"\n"
	for i in G.getEqNodes():
	    for j in i.getAllSaturatedEdgesInG():
		#print str(j.getDestination())+' -------- '+ \
		#	printer.dumpEq(i.eq)
		msg=msg+str(j.getDestination())+' -------- '+ \
			printer.dumpEq(i.eq)+"\n"
	return msg
