# MuModelica Compiler Project
# Author: Steven Xu
# Modeling, Simulation and Design Lab (MSDL)
# School of Computer Science
# McGill University

# This is the visitor module


import copy
from AST.ASTComponent import ClassModification
from AST.ASTComponent import EqualsModification, AssignModification
from AST.ASTComponent import ElementRedeclaration
from AST.ASTExp import * #IdentExp, SingleCompRefExp, CompRefExp
from AST.ASTEquations import RegularEqStm, ConnectEqStm
from FrontEnd.TypeChecker1 import TypeChecker
from FrontEnd.miscellaneous.CopyMaker import CopyMaker

# ---------------------------------------------------------------------
class Flattener:
    "flatten a model down to predefined types"
    def __init__(self, ast):
	# the AST (ClassFile)
	self.ast=ast
	# the target class
        #self.modelName=modelNameP
	self.tc=TypeChecker()
	self.copier=CopyMaker()
	self.currentPrefix=''

    def go(self):
	# get the main class
	#classDef=self.ast.getScope().localLookup(self.modelName).getDefinition()
	mainModel=self.ast.classFiles[0].classDefinitions[0]
	self.flattenClass(mainModel)
	return mainModel
	# dump flattened version declarations
	#classDef.dumpFlattenedVersion()
	#printer=FlatModelicaPrinter()
        #classDef.accept(printer)
	#return printer.getOutput()

    # classDef: AST::ClassDefinition
    def flattenClass(self, classDef):
	for i in classDef.getScope().getDeclarations():
	    # case 1: primitive components
	    if self.isPredefinedType(i): # or self.isBuiltinType(i):
	    	if i.hasModif():	
		    self.mergePredefinedModif(i)
		else:
		    i.createEnv()
		classDef.addFlattenedComponent(str(i), i)
	    # case 2: composite components
	    else:
		# flatten composite component: return a tuple of 3 lists:
		# list of declarations, list of initialEquations, and 
		# list of equations
		(dl, il, el)=self.flattenComponent(i)
		for i in dl:
		    classDef.addFlattenedComponent(str(i), i)
		classDef.flattenedInitEqs=classDef.flattenedInitEqs+il
		classDef.flattenedEqs=classDef.flattenedEqs+el
	# generate connection equations
	connectEqStmList=classDef.getConnectEqStms()
	if len(connectEqStmList)>0:
            classDef.flattenedEqs=classDef.flattenedEqs+\
				self.genConnectionEqs(connectEqStmList)
	# resolve names in equations
	self.currentPrefix=''
	for i in classDef.getInitEquations():
	    if isinstance(i, RegularEqStm):
	    	classDef.flattenedInitEqs.append(i)
	for i in classDef.flattenedInitEqs:
	    i.accept(self)
	for i in classDef.getEquations():
	    if isinstance(i, RegularEqStm):
	    	classDef.flattenedEqs.append(i)
	for i in classDef.flattenedEqs:
	    i.accept(self)
	# turn modificatons into equations
	for i in classDef.getFlattenedComponents():
	    eq=self.createModifEq(i)
	    if not eq==None:
	    	classDef.flattenedEqs.append(eq)
	# convert CompRefExp to IdentExp
        self.convertCompRefExpToIdentExp(classDef)

    # flatten a composite component
    # decl: ASTScope::DeclarationEntry
    def flattenComponent(self, decl):
	declList=[]
	initEqList=[]
	eqList=[]
	classDef=decl.getDefinition()
	# case 1: an instance has been created
	if decl.hasInstance():
	    # recursively flatten all components in this instance
	    for i in decl.getInstance().getDeclarations():
		if self.isBuiltinType(i):
		    pass
		elif self.isPredefinedType(i): # or self.isBuiltinType(i):
		    self.mergePredefinedModif(i)
		    declList.append(i)
		else:
		    (declList1, initEqList1, eqList1)=self.flattenComponent(i)
		    declList=declList+declList1
		    initEqList=initEqList+initEqList1
		    eqList=eqList+eqList1
	    initEqList=initEqList+decl.getInstance().getInitEqs()
	    eqList=eqList+decl.getInstance().getEqs()
	# case 2: no instance was created (no modification)
	else:
	    # recursively flatten all components in this instance
	    for i in classDef.getScope().getDeclarations():
		if self.isBuiltinType(i):
                    pass
		elif self.isPredefinedType(i): # or self.isBuiltinType(i):
		    j=i.deepCopy()
		    self.mergePredefinedModif(j)
		    declList.append(j)
		else:
		    (declList1, initEqList1, eqList1)=self.flattenComponent(i)
		    declList=declList+declList1
		    initEqList=initEqList+initEqList1
		    eqList=eqList+eqList1
	    for i in decl.getDefinition().getInitEquations():
		initEqList.append(self.copier.deepcopyEqStm(i))
	    for i in decl.getDefinition().getEquations():
		eqList.append(self.copier.deepcopyEqStm(i))
	# generate connection equations
        connectEqStmList=classDef.getConnectEqStms()
        if len(connectEqStmList)>0:
            eqList=eqList+self.genConnectionEqs(connectEqStmList)
	# remove connection eqs from the eq list
	index=len(eqList)-1
	while index >= 0:
	    if isinstance(eqList[index], ConnectEqStm):
		del eqList[index]
	    index=index-1
	# resolve names in declarations and equations
	for i in declList:
	    self.currentPrefix=str(decl.getName())
	    # resolve names in declarations
	    i.getName().setValue(self.currentPrefix+'_'+str(i.getName()))
	    # resolve names in the environment
	    env=i.getEnv()
	    # eg., in component x5(a=x3.a);
	    # Equaions x5_a=x3_a will be generated. Prefix 'x5' is not 
	    # applied to the RHS
	    self.currentPrefix=''
	    for i in env.keys():
		elem=env[i]
		if not elem==None:
		    elem.accept(self)
	self.currentPrefix=str(decl.getName())
	for i in initEqList:
	    i.accept(self)
	for i in eqList:
	    i.accept(self)
	return (declList, initEqList, eqList)


    # merge modifications at the level of predefined types
    def mergePredefinedModif(self, decl):
	decl.createEnv()
	for i in decl.getModif():
	    if isinstance(i, ClassModification):
		for j in i.getArgumentList():
		    if isinstance(j, ElementRedeclaration):
			raise 'Error: \"redeclare\" not yet supported!'
		    else: # ElementModification
			compRef=j.getCompRef()    
			if not decl.getEnv().has_key(str(compRef)):
			    raise 'Error: illegal modif of component'
			if isinstance(j.getModification(), ClassModification):
			    raise 'Error: illegal use of ClassModificatio'
			decl.addToEnv(str(compRef), \
				j.getModification().getExp())
		if not i.getEqualsModification()==None:
		    decl.addToEnv('value', \
			i.getEqualsModification().getExp())
	    elif isinstance(i, EqualsModification):
		decl.addToEnv('value', i.getExp())

    # Curently only eq for 'value' field is created
    # Other fields are ignored
    def createModifEq(self, decl):
	if decl.isParameter() or decl.isConstant():
	    return None
	env=decl.getEnv()
	# need create eqs for modif to all attributes, not only "value"
	if not env["value"]==None:
	    LHS=IdentExp(-1,-1, str(decl))
	    RHS=env["value"]
	    eq=RegularEqStm(-1, -1, LHS, RHS)
	    return eq
	return None
	
    # generate connection equations from connection statments
    # listOfConnectEqStm: [] of ASTEquations::ConnectEqStm
    def genConnectionEqs(self, listOfConnectEqStm):
	tc=TypeChecker()
	connectionEqs=[]
	listOfConnSet=[]
	# build connection sets
	for i in listOfConnectEqStm:
	    connector1=i.getFirstConnector()
	    connector2=i.getSecondConnector()
	    # type check that the 2 connectors are type identical
	    decl1=self.getConnectorInstance(connector1)
	    decl2=self.getConnectorInstance(connector2)
	    if not tc.isTypeIdentical(decl1, decl2):
		print "Connectors: "+str(connector1)+" and "+\
			str(connector2)+" are NOT type identical"
		raise "Connectors are not type identical!"
	    connSet=[connector1, connector2]
	    listOfConnSet.append(connSet)
	# merge connection sets
	listOfConnSet=self.mergeConnectionSet(listOfConnSet)
	# generate eqs
	for connSet in listOfConnSet:
	    inst=self.getConnectorInstance(connSet[0])
	    classDef=inst.getDefinition()
	    # verify that type prefix "flow" is applied to subtype of Real 
 	    # ...(need modify typeChecker)
	    # fatten the connector class
	    self.flattenClass(classDef)
	    for decl in classDef.getFlatScope().getDeclarations():
		if not decl.isFlow():
		    # generate equations for cross variables
		    # decl_0=decl_1, decl_0=decl_2, ... decl_0=decl_n
		    index=1
		    while index <len(connSet):
			# lhs: decl_0
			lhs=CompRefExp(connSet[0].getLineNum(), \
			    connSet[0].getColNum, \
			    connSet[0].getOperands()+\
			    [IdentExp(-1, -1, str(decl))])
			# rhs: decl_i
			rhs=CompRefExp(connSet[index].getLineNum(), \
                            connSet[index].getColNum, \
			    connSet[index].getOperands()+\
                            [IdentExp(-1, -1, str(decl))])
			eq=RegularEqStm(connSet[0].getLineNum(), \
			   connSet[0].getColNum, lhs, rhs)
			connectionEqs.append(eq)
			index=index+1
		else:
		    # generate eqs for flow vars
		    if self.isOutsideConnector(connSet[0]):
			lineNum=connSet[0].getLineNum()
			colNum=connSet[0].getColNum
			lhs=MulExp(lineNum, colNum, \
			    [RealExp(lineNum, colNum, -1.0), \
			    CompRefExp(lineNum, colNum,\
				connSet[0].getOperands()+\
				[IdentExp(-1, -1, str(decl))])])
		    else:
			lineNum=connSet[0].getLineNum()
			colNum=connSet[0].getColNum
			lhs=CompRefExp(lineNum, colNum,\
                                connSet[0].getOperands()+\
                                [IdentExp(-1, -1, str(decl))])
		    index=1
		    while index<len(connSet):
			lineNum=connSet[index].getLineNum()
			colNum=connSet[index].getColNum
			if self.isOutsideConnector(connSet[index]):
			    newOperand=MulExp(lineNum, colNum, \
				[RealExp(lineNum, colNum, -1.0), \
				CompRefExp(lineNum, colNum,\
                                connSet[index].getOperands()+\
                                [IdentExp(-1, -1, str(decl))])])
			else:
			    newOperand=CompRefExp(lineNum, colNum,\
                                connSet[index].getOperands()+\
                                [IdentExp(-1, -1, str(decl))])
			lhs=SumExp(lineNum, colNum, [lhs, newOperand])
			index=index+1
		    rhs=RealExp(lineNum, colNum, 0.0)
		    eq=RegularEqStm(lineNum, colNum, lhs, rhs)
		    connectionEqs.append(eq)
	return connectionEqs

    # merge connection sets
    def mergeConnectionSet(self, listOfConnSet):
	index=len(listOfConnSet)-1
	while index>0:
	  # 2 sets to be merged
	  firstSet=listOfConnSet[index]
	  index2=0
	  while index2<len(listOfConnSet):
	    #secondSet=listOfConnSet[index-1]
	    if index2!=index:
	        secondSet=listOfConnSet[index2]
	    else:
		index2=index2+1
		continue
	    # Felems: elems only appear in firstSet
	    # Selems: elems only appear in secondSet
	    # commentElems: elems appear in both sets 
	    Felems=[]
	    Selems=[]
	    commentElems=[]
	    # pick out comment elements and Felems
	    for i in firstSet:
		isComment=False
		for j in secondSet:
		    if str(i)==str(j):
			commentElems.append(i)
			isComment=True
			break
		if isComment:
		    continue
		else:
		    Felems.append(i)
	    # pick out Selems
	    for i in secondSet:
		isComment=False
		for j in commentElems:
		    if str(i)==str(j):
			isComment=True
			break
		if isComment:
                    continue
                else:
		    Selems.append(i)
	    # if the two sets have comment elements, then merge
	    if len(commentElems)>0:
		#listOfConnSet[index-1]=Felems+commentElems+Selems
		listOfConnSet[index2]=Felems+commentElems+Selems
		del listOfConnSet[index]
	    index2=index2+1
	  index=index-1
	return listOfConnSet

    # identify if a connector is an outside connector
    # conn: ASTExp::ConnectRefExp
    def isOutsideConnector(self, conn):
	operands=conn.getOperands()
	if len(operands)==2:
	    firstElem=operands[0] # ASTExp::SingleCompRefExp
	    classDef=firstElem.getOperands()[0].getValue().getDefinition()
	    if not classDef.isConnector():
		return False
	return True
	
    #-------------------auxiliary methods --------------------
    # get the connector instance 
    # connRefExp: ASTExp::ConnectRefExp
    # return: ASTScope::DeclarationnEntry
    def getConnectorInstance(self, connRefExp):
	operands=connRefExp.getOperands()
	if len(operands)==1:
	    decl=operands[0].getOperands()[0].getValue()
	else:
	    decl=operands[1].getOperands()[0].getValue()
	return decl

    # return string representation of ConnectRefExp with 
    # dot referencing replaceed by underscore,
    # eg, a.b -> a_b
    def toStringWoDotRef(self, connRef):
	result=''
        if len(connRef.operands)>0:
            result=str(connRef.operands[0])
        if len(connRef.operands)==2:
            result=result+'_'+str(connRef.operands[1])
        return result

    # elem: DeclarationEntry
    def isPredefinedType(self, elem):
	predefinedScope=self.ast.predefined.classDefinitions[0].scope
        found=predefinedScope.localLookup(str(elem.getType()))
        if not found==None:
            return True
        else:
            RealDefEntry=predefinedScope.localLookup('Real')
            RealDef=RealDefEntry.getDefinition()
            IntegerDefEntry=predefinedScope.localLookup('Integer')
            IntegerDef=IntegerDefEntry.getDefinition()
            StringDefEntry=predefinedScope.localLookup('String')
            StringDef=StringDefEntry.getDefinition()
            BooleanDefEntry=predefinedScope.localLookup('Boolean')
            BooleanDef=BooleanDefEntry.getDefinition()
            elemDef=elem.getDefinition()
            if self.tc.isClassTypeEquivalent(elemDef, RealDef) or \
                self.tc.isClassTypeEquivalent(elemDef, IntegerDef) or \
                self.tc.isClassTypeEquivalent(elemDef, StringDef) or \
                self.tc.isClassTypeEquivalent(elemDef, BooleanDef):
                return True
        return False

    def isBuiltinType(self, elem):
	found=self.ast.builtinScope.localLookup(str(elem.getType()))
	return (not found==None)

    
    # ---------------------resolving names in equations -----------
    def visitRegularEqStm(self, eqStm):
        eqStm.LHS.accept(self)
        eqStm.RHS.accept(self)

    def visitConnectEqStm(self, eqStm):
        eqStm.first.accept(self)
        eqStm.second.accept(self)

    def visitFunCallEqStm(self, eqStm):
        eqStm.funCall.accept(self)

    def visitConditionalEqStm(self, eqStm):
        eqStm.ifTest.accept(self)
        for i in eqStm.ifBody:
            i.accept(self)
        for i in eqStm.elseifList:
            i.accept(self)
        for i in eqStm.elseBody:
            i.accept(self)

    def visitForEqStm(self, eqStm):
        eqStm.forIndices.accept(self)
        # enter new scope
        #self.currentScope=eqStm.getScope()
        for i in eqStm.body:
            i.accept(self)
        # reset current scope
        #self.currentScope=self.currentScope.getOuter()

    def visitWhenEqStm(self, eqStm):
        eqStm.whenTest.accept(self)
        for i in eqStm.body:
            i.accept(self)
        for i in elsewhenList:
            i.accept(self)

    # ------------------exp------------------
    def visitIdentExp(self, exp):
	if not self.currentPrefix=='':
	    if not str(exp)=='time':
	    	exp.setValue(self.currentPrefix+'_'+str(exp))

    def visitRealExp(self, exp):
        pass

    def visitIntegerExp(self, exp):
        pass

    def visitBoolExp(self, exp):
        pass

    def visitStringExp(self, exp):
        pass

    # this method is not visited
    def visitNameExp(self, exp):
	pass

    def visitOrExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitAndExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitNotExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitLessExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def vistLessEQExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitGreaterExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitGreaterEQExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitEQEQExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitLessGtExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitSumExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitMulExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitSubtractExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitDivExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitUnaryMinusExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitPowerExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitRangeExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitIfElseExp(self, exp):
        exp.ifTest.accept(self)
        exp.ifBody.accept(self)
        for i in exp.elseifList:
            i.accept(self)
        if not exp.elseBody==None:
            exp.elseBody.accept(self)

    def visitFunCallExp(self, exp):
	operands=exp.getOperands()
        # take no action on function name
        for i in range(1, len(operands)):
            operands[i].accept(self)

    def visitFunArgsExp(self, exp):
	for i in exp.getOperands():
            i.accept(self)

    def visitForFunArgExp(self, exp):
        #operands=exp.getOperands()
        # enter new scope
        #self.currentScope=exp.getScope()
        # only the first elem is looked up in the innner scope
        #operands[0].accept(self)
        # exit scope
        #self.currentScope=self.currentScope.getOuter()
        #for i in range(1, len(operands)):
        #    operands[i].accept(self)
	pass # need to resolve

    def visitNamedArgExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitConnectRefExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitSingleCompRefExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitCompRefExp(self, exp):
	originalName=''
	operands=exp.getOperands()
	i=len(operands)-1
	while i>=0:
	    originalName='_'+str(operands[i])+originalName
	    if not i==0:
		del exp.operands[i]
	    i=i-1
	if self.currentPrefix=='':
	    originalName=originalName[1:len(originalName)]
	id=IdentExp(-1, -1, self.currentPrefix+originalName)
	exp.operands[0]=SingleCompRefExp(-1, -1, [id])

    def visitArrSubExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitCommentExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    def visitAnnotationExp(self, exp):
        pass

    def visitIdentInExpExp(self, exp):
        exp.getOperands()[1].accept(self)

    def visitForIndicesExp(self, exp):
        for i in exp.getOperands():
            # i: IdentInExpExp
            i.accept(self)

    def visitExpListExp(self, exp):
        for i in exp.getOperands():
            i.accept(self)

    # -----------------------------------------------------------
    # convert all CompRefExp to IdentExp in classDef
    # classDef: AST::ClassDefinition
    def convertCompRefExpToIdentExp(self, classDef):
	#mainEntry=self.ast.getScope().localLookup(self.modelName)
        #mainModel=mainEntry.getDefinition()
        for i in classDef.getFlattenedInitEqs():
            self.convertCompRefInEq(i)
        for i in classDef.getFlattenedEqs():
            self.convertCompRefInEq(i)

    def convertCompRefInEq(self, eq):
        if isinstance(eq, RegularEqStm):
            eq.setLHS(self.convertCompRef(eq.getLHS()))
            eq.setRHS(self.convertCompRef(eq.getRHS()))
	else:
            raise 'Currently only handles regular equations in \
                        canonical transformation!'

    def convertCompRef(self, exp):
        if isinstance(exp, CompRefExp):
            return IdentExp(exp.getLineNum(), exp.getColNum(), str(exp))
	elif isinstance(exp, IdentExp):
	    return exp
        elif isinstance(exp, SumExp) or \
	     isinstance(exp, SubtractExp) or \
             isinstance(exp, MulExp) or \
             isinstance(exp, DivExp) or \
             isinstance(exp, PowerExp) or \
	     isinstance(exp, UnaryMinusExp):
            operands=exp.getOperands()
            index=0
            for index in range(0, len(operands)):
                exp.operands[index]=self.convertCompRef(operands[index])
            return exp
	# need loop over all Exp instances here, eg, FunCallExp ...
        else:  
            return exp

