
package queue
  import devs.*;
  import externalfunctions.*;
  import simulator.*;

  class GeneratorState
    Generator.SeqStates seqState(start=Generator.SeqStates.G_IDLE);
  end GeneratorState;

  class Generator
    extends AtomicDEVS;
    parameter Integer ia=0;
    parameter Integer ib=0;
    parameter Integer szl=0;
    parameter Integer szh=0;
    parameter String name="a";
    output DevsPort g_out;
    GeneratorState state();
    type SeqStates = enumeration(G_IDLE, G_GENERATING);

    function intTransition
    algorithm
      if( state.seqState == SeqStates.G_IDLE ) then
        state.seqState := SeqStates.G_GENERATING;
      elseif( state.seqState == SeqStates.G_GENERATING ) then
        state.seqState := SeqStates.G_IDLE;
      end if;
    end intTransition;

    function outputFnc
      DevsEvent evt = null;
    algorithm
      if( state.seqState == SeqStates.G_GENERATING) then
        evt := Job(szl, szh);
        poke(g_out, evt);
      end if;
    end outputFnc;

    function timeAdvance
      output Integer timespan;
    algorithm
      if( state.seqState == SeqStates.G_IDLE) then
        timespan:=randint(ia, ib);
      end if;
      if( state.seqState == SeqStates.G_GENERATING) then
        timespan := 0;
      end if;
    end timeAdvance;

  end Generator;

  class ProcessorState
    Job currentJob = null;
    list queue();
    Integer queueSize = 0;
    Real timeElapsed = 0.0;
    Processor.SeqStates seqState(start=Processor.SeqStates.P_IDLE);
  end ProcessorState;

  class Processor
    extends AtomicDEVS;
    parameter Integer qSize=0;
    parameter String name="a";
    input DevsPort p_in;
    output DevsPort p_out;
    output DevsPort p_discard;
    ProcessorState state();
    type SeqStates = enumeration(P_IDLE, P_BUSY, P_DISCARD);

    function initialization
    algorithm
      state.queueSize := qSize;
    end initialization;


    function extTransition
    algorithm
      if( state.seqState == SeqStates.P_IDLE ) then
        state.currentJob:=peek(p_in);
        state.timeElapsed:=0;
        state.seqState := SeqStates.P_BUSY;
      elseif( state.seqState == SeqStates.P_BUSY ) then
        state.timeElapsed:=state.timeElapsed+elapsed;
        if( len(state.queue)<state.queueSize ) then
          state.queue.append(peek(p_in));
          state.seqState := SeqStates.P_BUSY;
        elseif( len(state.queue)==state.queueSize ) then
          state.timeElapsed:=state.timeElapsed + elapsed;
          state.seqState := SeqStates.P_DISCARD;
        end if;
      end if;
    end extTransition;


    function intTransition
    algorithm
      if( state.seqState == SeqStates.P_BUSY ) then
        if( len(state.queue)==0 ) then
          state.seqState := SeqStates.P_IDLE;
        elseif( len(state.queue)>0 ) then
          state.timeElapsed:=0;
          state.currentJob:=state.queue.pop();
          state.seqState := SeqStates.P_BUSY;
        end if;
      elseif( state.seqState == SeqStates.P_DISCARD ) then
        state.seqState := SeqStates.P_BUSY;
      end if;
    end intTransition;

    function outputFnc
      DevsEvent evt = null;
    algorithm
      if( state.seqState == SeqStates.P_BUSY) then
        poke(p_out, state.currentJob);
        state.currentJob:=null;
      end if;
      if( state.seqState == SeqStates.P_DISCARD) then
        poke(p_discard, peek(p_in));
      end if;
    end outputFnc;

    function timeAdvance
      output Integer timespan;
    algorithm
      if( state.seqState == SeqStates.P_IDLE) then
        timespan:=INFINITY;
      end if;
      if( state.seqState == SeqStates.P_BUSY) then
        timespan := state.currentJob.size - state.timeElapsed;
      end if;
      if( state.seqState == SeqStates.P_DISCARD) then
        timespan:=0;
      end if;
    end timeAdvance;

  end Processor;

  class Processors
    extends CoupledDEVS;
    parameter Integer rsize;
    parameter String name;
    output DevsPort r_out;
    output DevsPort r_discard;
    input DevsPort r_in;
    Processor p1(qSize=rsize);
    Processor p2(qSize=rsize);
    Processor p3(qSize=rsize);
  equation
    connect( p1.p_out, r_out);
    connect( p1.p_discard, p2.p_in);
    connect( p2.p_out, r_out);
    connect( p2.p_discard, p3.p_in);
    connect( p3.p_out, r_out);
    connect( p3.p_discard, r_discard);
    connect( r_in, p1.p_in);
  end Processors;

  class SimModel
    extends CoupledDEVS;
    parameter Integer sia;
    parameter Integer sib;
    parameter Integer sszl;
    parameter Integer sszh;
    parameter Integer szise;
    parameter String name;
    Generator g1(ia=sia, ib=sib, szl=sszl, szh=sszh);
    Processors tp1(rsize=ssize);
  equation
    connect( g1.g_out, tp1.r_in);
  end SimModel;

  class Job
    extends DevsEvent;
    parameter Integer szl;
    parameter Integer szh;
    Integer id = 0 "ID" ;
    Integer size = randint(szl, szh);
  end Job;

  class ProcessorsExperiment
    extends DevsExperiment;
    SimModel simmodelModel(sia=2, sib=5, sszl=3, sszh=10, ssize=2, name="ProcessorsExperiment");
    Simulator sim(simModel=simmodelModel);
  end ProcessorsExperiment;

end queue;