from GraphRepr.GraphElement import GraphElement
from GraphRepr.Node import Node


from GraphRepr.Error.GraphException import\
                                InvalidPathException,\
                                NonUniqueIdException,\
                                InvalidIdException,\
                                NotConnectedException,\
                                AlreadyConnectedException,\
                                InvalidElementException


#from GraphRepr.GraphUtils.GraphListCollection import GraphListCollection
from GraphRepr.GraphUtils.GraphElemCollection import GraphElemCollection
from GraphRepr.GraphUtils.GraphConnCollection import GraphConnCollection

from string import replace
import weakref

# Class that supports HiGraphs

class Graph(GraphElement):

    # Constructor for Graph.  Each HiGraph has a root
    # that manages the connections (for efficiency reasons).
    # If no parent is specified, the instantiated graph will
    # act as the root.   A type can be specifed that will be used
    # by a typeChecker for validation. If
    # no type is specified, the typechecker will use the same graph to
    # validate.
    # Each graph has the following private attributes:
    # Two mappings for the element that the graph contains:
    #  1) graphElements_ID == localId -> element
    #  2) graphElements == type -> list of element having this type
    #  About IDs, two main concept:
    #  A) localId : id of an element unique with respect to a certain
    #               parent, which is not the root.
    #               (e.g. ER.entity, state, entity, etc)
    #  B) globalId: full path from the root to the element.
    #               GLOBALLY unique to the whole graph.
    #               (e.g. root.ER.entity, root.FSA.state, etc)
    #  Each subgraph has a reference to the root
    #  (to have access to its connection)
    #  The root has a mapping for the connections:
    #  elemConn == globalId -> ([IN connections mapping, OUT connections mapping])
    #  where IN and OUT connections mapping are:
    #  elemID -> elemRef
    #  e.g.  Entity is connected to Entity, we will have something like:
    #  elemConn == {"Entity"::[{"Entity"::EntityRef}, {"Entity"::EntityRef}]}
    #  Each graph also has a reference to its direct parent and a depth value.

    # IN_A   : ID::str. A local id for this graph that will be unique within
    #                   the scope of the parent
    # [IN_A] : parent::Graph. An arbitrary parent for this graph. If not specified,
    #                         this graph will act as a root
    # [IN_B] : type::str. A potential type for this (sub)graph. Local id located in
    #                     the scope of the type of the parent of this graph
    # RETURN : this object..
    # ORDER: O(1)

    def __init__(self, ID, parent=None, type=None):
        #basic attributes.  (not inherited from GraphElement : private..)
        self.__id=ID

        if type==None:
            self.__typeId=ID
        else:
            self.__typeId=type

        self.__type=None

        #dictionnaries for elements, indexed by type
        self.__graphElements={}

        #dictionnaries for elements, indexed by id
        self.__graphElements_ID={}


        self.__parent=None
        self.__root=None
        self.__globalId=str(ID)
        self.__elemConn={}
        self.__depth=0

        if parent != None:
            #We add ourself to our parent
            parent.addElement(self)

    def getElemTypes(self):
        return self.__graphElements.keys()

    #Function called when an element is popped out of this graph.
    #Create a new context for the removed element :
    #  If the element is a graph: it becomes a root itself
    #  If the element is a node: its root is set to none
    #Remove the connections that are not
    #meaningful to the new context of the removed graph.
    #i.e. keep only the connections where BOTH members
    #are within the scope of the removed graph. A node that is removed
    #loses all its connections since it is not member of any graph
    #and contains no element. The function also delete all the information
    #that this graph has about the removed element.
    #IN: LocalID (str)
    #RETURNS: GraphElement
    #ORDER: O(n+e)   n: elements, e: edges

    def updatePopElem(self, ID): #PROTECTED

        #getting the elements of this graph
        elem=self.__graphElements_ID[ID]

        #current global Id
        oldGlobal=elem.getGlobalId()

        if isinstance(elem,Graph):
            #if it is a graph, we need to keep valid connections
            #in the context where this graph is becoming a root
            #itself

            #getting the mapping of all the connections for the whole HiGraph
            conns = self.getRoot().getConnInfo()
            #will append the connections to keep
            newConns = []

            #looping over all connections to decide which ones to keep
            for key in conns.keys():
                #key is an ID of a given element of the whole HIGRAPH
                #if key is within the scope of "elem", we may keep it.
                #i.e.: the globalId of elem (oldGlobal) is contained within the
                #key
                if key[0:(len(oldGlobal))]==oldGlobal and oldGlobal!=key:
                    for inKey in conns[key][0].keys():
                        #We now need to verify that the other endpoint of the
                        #connection is within the scope of elem. Same process.
                        if inKey[0:(len(oldGlobal))]==oldGlobal and\
                          oldGlobal!=inKey:
                            #since BOTH endpoints of a given connection are
                            #contained in "elem", we keep the connection
                            inElem = conns[key][0][inKey][0]
                            outElem = conns[inKey][1][key][0]
                            #we append a valid connection for the new context.
                            newConns.append((inElem, outElem))


            #we remove all the connection information known by this higraph
            #related to elem (i.e. elem is removed from this graph, it has no
            #more connections in THIS graph)
            elem.disconnectAllRecurse()

            #Elem becomes itself a root, we set it to None.
            elem.setRoot(None)

            #The globalId of elem becomes its localId
            elem.setGlobalId(elem.getLocalId())

            #Elem is a root, it has no parent.
            elem.setParent(None)

            #Elem is a root, its depth is 0
            elem.setDepth(0)

            #We create the datastructure that will manages the connections
            #within elem, since elem is now a root, it will manages
            #the connections
            elem.createConnInfo()

            #We connect the valid connections in the new context of elem.
            for i,o in newConns:
                elem.connect(i.getGlobalId(1), o.getGlobalId(1))
        else:
            #Node case, same as Graph, but without connections.

            elem.disconnectAllRecurse()
            elem.setRoot(None)
            elem.setGlobalId(elem.getLocalId())
            elem.setParent(None)
            elem.setDepth(0)

        #Removing "elem" from the graph.  It is no longer a member
        #of "self", but a stand-alone root.
        elemList=self.__graphElements[elem.getTypeLocalId()]
        elemList.remove(elem)
        del self.__graphElements_ID[ID]

        return elem

    #Message sent to the parent when one of its child is removed
    #Deletes all the known information that the parent knows about the element
    #and its connections information located in the root
    #IN: ID::str localId of the removed element with respect to the parent
    #RETURN: None
    #Order: O(elements+edges)

    def updateRemoveElem(self, ID): #PROTECTED
        elem=self.__graphElements_ID[ID]

        elem.disconnectAllRecurse()
        elemList=self.__graphElements[elem.getTypeLocalId()]
        elemList.remove(elem)
        del self.__graphElements_ID[ID]

    #Message sent to the parent of a newly added element to
    #update the datastructes.
    #IN: elem, reference to a GraphElement
    #Order: O(1)

    def setAddElement(self, elem):
        #mapping by ID..
        self.__graphElements_ID[elem.getLocalId()]=elem

        #add the actual elem to the graph
        if self.__graphElements.has_key(elem.getTypeLocalId()):
            elemList = self.__graphElements[elem.getTypeLocalId()]
            elemList.append(elem)
        else:
            newlist=[elem]
            self.__graphElements[elem.getTypeLocalId()]=newlist


    #Function called to set the GlobalId of this element.
    #Since the GlobalId of a child depends on his parent,
    #the function recursively calls itself for every child
    #This message is sent to an element when its context change
    #for example, if the element is popped out of a graph or added
    #in a new graph
    #IN: path (new global id to be set)
    #RETURN: None
    #ORDER: O(n) where n is the number of elements
    # Num of recursive calls: n

    def setGlobalId(self, path): #PROTECTED
        self.__globalId=path
        for key in self.__graphElements_ID.keys():
            elem=self.__graphElements_ID[key]
            newPath=path+"."+elem.getLocalId()
            elem.setGlobalId(newPath)

    #Function called to set the depth of this element.
    #Since the depth of a child depends on his parent,
    #the function recursively calls itself for every child
    #This message is sent to an element when its context change
    #for example, if the element is popped out of a graph or added
    #in a new graph
    #IN: depth (new depth to be set)
    #RETURN: None
    #ORDER: O(n) where n is the number of elements
    # Num of recursive calls: n

    def setDepth(self, depth): #PROTECTED
        self.__depth=depth
        for key in self.__graphElements_ID.keys():
            self.__graphElements_ID[key].setDepth(depth+1)

    #Function called to set the root of this element.
    #Since the root of a child depends on his parent,
    #the function recursively calls itself for every child
    #This message is sent to an element when its context change
    #for example, if the element is popped out of a graph or added
    #in a new graph
    #IN: root (new root to be set)
    #RETURN: None
    #ORDER: O(n) where n is the number of elements
    # Num of recursive calls: n

    def setRoot(self, root=None): #PROTECTED
        if root==None:
            self.__root=root
            for key in self.__graphElements_ID.keys():
                self.__graphElements_ID[key].setRoot(self)
        else:
            self.__root=weakref.ref(root)
            for key in self.__graphElements_ID.keys():
                self.__graphElements_ID[key].setRoot(root)

    #Function called to set the parent of this element.
    #This message is sent to an element when its context change
    #for example, if the element is popped out of a graph or added
    #in a new graph
    #IN: parent (new parent to be set)
    #RETURN: None
    #ORDER: O(1)
    # Num of recursive calls:1

    def setParent(self, parent): #PROTECTED
        if parent==None:
            self.__parent=parent
        else:
            self.__parent=weakref.ref(parent)

    #Function called to create the connection mapping
    #located at the root level.  The function creates
    #the mapping information for "this" element and
    #for each of its child.
    #RETURN: None
    #ORDER: 0(n)
    #num of recursive calls: n

    def createConnInfo(self): #PROTECTED
        conns=self.getRoot().getConnInfo()
        if self.getParent()!=None:
            conns[self.getGlobalId()]=[{}, {}]
        for key in self.__graphElements_ID.keys():
            elem=self.__graphElements_ID[key]
            elem.createConnInfo()

    #Function to have access to the connection mapping of a root
    #RETURN: connection mapping (see its full description in the constructor)
    #RAISES NotImplementedError if called not from the root
    #ORDER: O(1)

    def getConnInfo(self): #PROTECTED
        if self.__root != None:
            raise NotImplementedError("Method only valid at root level")
        return self.__elemConn

    #Function called when two elements are connected (via "connect")
    #to store the connection information in the root.
    # Suppose we have the following connection fromElement -> toElement
    # IN_1: reference to the "from" element of the connection
    # IN_2: reference to the "to" element of the connection
    # OUT: None
    # RAISES: AlreadyConnectedException
    # ORDER: O(1)

    def setConnInfo(self, fromElem, toElem, fromOrder=0, toOrder=0): #PROTECTED

        #Suppose we have the following connection fromElement -> toElement
        conns=self.getRoot().getConnInfo()

        # Getting the globalIds of the elements.
        fromId=fromElem.getGlobalId()
        toId=toElem.getGlobalId()

        if conns[fromId][1].has_key(toId) or\
           conns[toId][0].has_key(fromId):
            raise AlreadyConnectedException(fromId, toId)

        #Add "toElement" to the mapping of OUT CONNECTIONS of "fromElement"
        conns[fromId][1][toId]= (toElem, fromOrder)

        #Add "fromElement" to the mapping of IN CONNECTIONS of "toElement"
        conns[toId][0][fromId]= (fromElem, toOrder)

    #Function called when two elements are disconnected (via "disconnect")
    #to remove the connection information in the root.
    # Suppose we have the following connection fromElement -> toElement
    # IN_1: reference to the "from" element of the connection
    # IN_2: reference to the "to" element of the connection
    # OUT: None
    # RAISES: AlreadyConnectedException
    # ORDER: O(e) (O(1) on average)

    def unsetConnInfo(self, fromElem, toElem): #PROTECTED

        conns=self.getRoot().getConnInfo()

        fromId=fromElem.getGlobalId()
        toId=toElem.getGlobalId()

        if not conns[fromId][1].has_key(toId) or\
           not conns[toId][0].has_key(fromId):
            raise NotConnedException(fromId,
                                        toId)

        del conns[fromId][1][toId]
        del conns[toId][0][fromId]

    # Function used to disconnect every child of elem, recursively.
    # This function will remove a connection if one of its component
    # is in the hierarchy of elem.
    # This message is send to an element when it is removed from a graph
    # so that the information about its connections can be removed from the
    # root.
    # RETURN: None
    # Order : O(elements+edges)

    def disconnectAllRecurse(self): #PROTECTED
        conns = self.getRoot().getConnInfo()
        globalId=self.getGlobalId()

        for key in self.__graphElements_ID.keys():
            cur=self.__graphElements_ID[key]
            cur.disconnectAllRecurse()

        for fromKey in conns[globalId][0].keys():
            fromElem=conns[globalId][0][fromKey][0]
            fromElemId=fromElem.getGlobalId()
            del conns[fromElemId][1][globalId]
            if globalId==fromElemId:
                del conns[fromElemId][0][globalId]

        for toKey in conns[globalId][1].keys():
            toElem=conns[globalId][1][toKey][0]
            toElemId=toElem.getGlobalId()
            del conns[toElemId][0][globalId]
        del conns[globalId]

    #The following methods are accessible members of the Graph interface.
    #They can be used anywhere, anytime :)

    #Set the type of this element, given a localId located
    #in the typeGraph of the parent of this element
    #IN: newTypeId:str, localId
    #RETURN: None
    #Order O(1)

    def setType(self, newTypeId):
        self.__typeId=newTypeId
        self.__type=None

    #Function used to get the value of the local id of the type.
    #This local ID is relative to the typeGraph of the parent of
    #this element
    #RETURN str
    #Order: O(1)

    def getTypeLocalId(self):
        if self.__typeId==None:
            return self.getLocalId()
        return self.__typeId

    #Function that return the type graph of this element.
    #RETURN: Graph
    #Order: O(depth) the first time, O(1) afterwards

    def getType(self):
        if self.__type!=None:
            return self.__type()
        if self.getParent()==None:
            return self
        typeGraphId=self.getParent().getType().getGlobalId()
        TYPE=self.getRoot().getElementGlobal(typeGraphId+"."+self.__typeId)
        self.__type=weakref.ref(TYPE)
        return self.__type()

    #Function used to get the value of the global id of the type.
    #RETURN str
    #Order: O(1)

    def getTypeGlobalId(self):
        if self.getParent()==None:
            return self.__id
        return self.getParent().getType().getGlobalId()+"."+self.__typeId

    #Get the connections located within this element in the form of a list
    #of tuples ( (FROM, TO, FROM_ORDERING,TO_ORDERING), ...)
    #RETURN list (tuple (GraphElement,GraphElement,int,int))
    #Order O(elements+edges)

    def getConnTuples(self):
        answer=[]
        #mapping holding connection info
        conns = self.getRoot().getConnInfo()
        #iterating over all elements of this graph
        elemIter = self.getElements().iterate()
        while not elemIter.isDone():
            elem=elemIter.currentElement()
            elemIter.next()
            #getting the inConnections of "elem"
            elemInCs=conns[elem.getGlobalId()][0]
            for key in elemInCs.keys():
                elemInC=elemInCs[key][0]
                elemOrd=elemInCs[key][1]
                elemInCOrd=conns[key][1][elem.getGlobalId()][1]
                #appending each connection to the list of tuple
                answer.append( (elemInC, elem, elemInCOrd, elemOrd) )
            if isinstance(elem,Graph):
                #if elem is a graph, must recursively call getConnTuples()
                elemConns=elem.getConnTuples()
                for conn in elemConns:
                    answer.append(conn)
        return answer

    # Get a subset of the connections contained within this element in the
    # form of a list of tuples ( (FROM,TO,FROM_ORDERING,TO_ORDERING), ..).
    # The subset will
    # contain the connections where the common parent of its components is
    # THIS element.  So, this connection cannot be expressed at a lower
    # depth than the depth of this element.
    # RETURN: list (tuple (GraphElement,GraphElement,int,int))
    # Order: O(edges)

    def getLocalConn(self):
        conns=self.getRoot().getConnInfo()
        globalId=self.getGlobalId()
        answer=[]

        for key in conns.keys():
            if key[0:len(globalId)]==globalId and globalId != key:
                conn=conns[key]
                for inCkey in conn[0].keys():
                    inC=conn[0][inCkey][0]
                    outCOrd=conn[0][inCkey][1]

                    if inC.getDepth()<=self.getDepth():
                        continue

                    outC=conns[inCkey][1][key][0]
                    inCOrd=conns[inCkey][1][key][1]

                    inGlobal=inC.getGlobalId(self.getDepth())
                    outGlobal=outC.getGlobalId(self.getDepth())

                    from string import split
                    commonParentIN=split(inGlobal, ".")
                    commonParentOUT=split(outGlobal, ".")

                    if commonParentIN[0]==commonParentOUT[0]:
                        if commonParentIN[1]!=commonParentOUT[1]:
                            answer.append((inC, outC, inCOrd, outCOrd))
                        else:
                            if len(commonParentIN)==2 or len(commonParentOUT)==2:
                                answer.append((inC, outC, inCOrd, outCOrd))
        return answer

    #Function used to get the depth value of this graph
    #RETURN int
    #Order: O(1)

    def getDepth(self):
        return self.__depth

    #Returns true if this graph is the root
    #RETURN int
    #Order: O(1)

    def isRoot(self):
        return self.__root==None

    #Returns true if globalId is located within the whole
    #HiGraph
    #RETURN int
    #Order: O(1)


    def hasGlobalId(self, globalId):
        conns=self.getRoot().getConnInfo()
        return conns.has_key(globalId)

    #Returns a globalId, potentially "local" to a certain
    #parent. For example, given "root.ER.entity":
    #getGlobalId() == "root.ER.entity" (global)
    #getGlobalId(1) == "ER.entity"  (local to "root")
    #getGlobalId(2) == "entity"  (local to "ER")
    #IN: int, id will be local to this depth
    #RETURN: str, ID
    #Order: O(depth)

    def getGlobalId(self, relativeToDepth=0):
        if not relativeToDepth:
            return self.__globalId
        if relativeToDepth>self.getDepth():
            relativeToDepth=self.getDepth()
        ind=0
        i=0
        for j in range(len(self.__globalId)):
            if self.__globalId[j]==".":
                ind=j
                i+=1
            if i==relativeToDepth:
                break
        return self.__globalId[(ind+1):]

    #Returns the root of the graph in which this graph is
    #RETURN: Graph
    #Order: O(1)

    def getRoot(self):
        if self.__root==None:
            return self
        return self.__root()

    #Returns the parent of this graph
    #RETURN: Graph
    #Order: O(1)

    def getParent(self):
        if self.__parent==None:
            return self.__parent
        else:
            return self.__parent()

    #Given two local paths (starting from this graph),
    #connect two elements.  Suppose we have the following:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # this is ok:  root.connect("ER.Entity", "ER.Entity")
    #              root.getElement("ER").connect("Entity", "Entity"))
    # IN_A: localID (wrt this graph) of the "from" element of the connection
    # IN_B: localID (wrt this graph) of the "to" element of the connection
    # RETURNS: None
    # Order: O(len(localID)), or O(depth)

    def connect(self, fromPath, toPath, fromOrder=0, toOrder=0):
        from string import split
        fromPath=split(fromPath, ".")
        toPath=split(toPath, ".")

        #starting from this graph, and will go down until we reach
        #"from" and "to"
        fromElement=self
        toElement=self

        for elemId in fromPath:
            if isinstance(fromElement,Node):
                raise InvalidPathException(fromElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            fromElement=fromElement.getElement(elemId)

        for elemId in toPath:
            if isinstance(toElement,Node):
                raise InvalidPathException(toElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            toElement=toElement.getElement(elemId)

        self.getRoot().setConnInfo(fromElement,toElement, fromOrder, toOrder)

    #Connects two elements without having to use ID relative to a parent.
    #i.e. globalID unique for the whole system
    # IN_A: globalID of the "from" element of the connection
    # IN_B: globalID of the "to" element of the connection
    # Order: O(depth)

    def connectGlobal(self, fromPath, toPath, fromOrder=0, toOrder=0):

        from string import split
        fromPath=split(fromPath, ".")
        fromPath=fromPath[1:]

        toPath=split(toPath, ".")
        toPath=toPath[1:]

        fromElement=self.getRoot()
        toElement=self.getRoot()

        for elemId in fromPath:
            if isinstance(fromElement,Node):
                raise InvalidPathException(fromElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            fromElement=fromElement.getElement(elemId)

        for elemId in toPath:
            if isinstance(toElement,Node):
                raise InvalidPathException(toElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            toElement=toElement.getElement(elemId)

        self.getRoot().setConnInfo(fromElement,toElement,fromOrder,toOrder)

    #Given two local paths (starting from this graph),
    #disconnect two elements.  Suppose we have the following:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # this is ok:  root.disconnect("ER.Entity", "ER.Entity")
    #              root.getElement("ER").disconnect("Entity", "Entity"))
    # IN_A: localID (wrt this graph) of the "from" element of the connection
    # IN_B: localID (wrt this graph) of the "to" element of the connection
    # RETURNS: None
    # Order: O(len(localID)), or O(depth)

    def disconnect(self, fromPath, toPath):
        from string import split
        fromPath=split(fromPath, ".")
        toPath=split(toPath, ".")

        fromElement=self
        toElement=self

        for elemId in fromPath:
            if isinstance(fromElement,Node):
                raise InvalidPathException(fromElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            fromElement=fromElement.getElement(elemId)

        for elemId in toPath:
            if isinstance(toElement,Node):
                raise InvalidPathException(toElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            toElement=toElement.getElement(elemId)

        self.getRoot().unsetConnInfo(fromElement,toElement)

    #Disconnects two elements without having to use ID relative to a parent.
    #i.e. globalID unique for the whole system
    # IN_A: globalID of the "from" element of the connection
    # IN_B: globalID of the "to" element of the connection
    # Order: O(depth)

    def disconnectGlobal(self, fromPath, toPath):

        from string import split
        fromPath=split(fromPath, ".")
        fromPath=fromPath[1:]

        toPath=split(toPath, ".")
        toPath=toPath[1:]

        fromElement=self.getRoot()
        toElement=self.getRoot()

        for elemId in fromPath:
            if isinstance(fromElement,Node):
                raise InvalidPathException(fromElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            fromElement=fromElement.getElement(elemId)

        for elemId in toPath:
            if isinstance(toElement,Node):
                raise InvalidPathException(toElement.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            toElement=toElement.getElement(elemId)

        self.getRoot().unsetConnInfo(fromElement,toElement)

    #Returns the localId of this element with respect to its parent
    #RETURNS: str
    #Order O(1)

    def getLocalId(self, relativeToParent=0):
        if not relativeToParent:
            return self.__id
        if relativeToParent >= self.getDepth():
            return self.__globalId
        answer=self.__id
        curParent=self
        for i in range(relativeToParent):
            curParent=curParent.getParent()
            answer = curParent.getLocalId()+"."+answer
        return answer

    #Returns true if ID is present within this graph.  ID is "local" with
    #respect to this graph.  e.g.:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # root.hasLocalId("ER") == 1  (root is "this" graph)
    # root.hasLocalId("ER.Entity") == 1
    # root.hasLocalId("Entity") == 0
    # root.getElement("ER").hasLocalId("Entity") == 1  (now ER is "this" graph)
    # etc
    # IN: ID::str
    # RETURN: boolean
    # Order: O(depth)


    def hasLocalId(self, ID):
        from string import split
        IDPath=split(ID, ".")

        if len(IDPath)==1:
            return self.__graphElements_ID.has_key(IDPath[0])

        parent=self

        for elemId in IDPath:
            if isinstance(parent,Node):
                raise InvalidPathException(parent.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            if parent.hasLocalId(elemId):
                parent=parent.getElement(elemId)
            else:
                return 0

        return 1

    def getRootElem(self):
        import random,time
        random.seed(time.time())
        #random.randint(0,len(self.__graphElements_ID.keys())-1)
        #ID=self.__graphElements_ID.keys()[0]
        #T=0
        #if T<len(self.__graphElements_ID.keys()):
        #    ID=self.__graphElements_ID.keys()[T]
        ID=self.__graphElements_ID.keys()[random.randint(0,len(self.__graphElements_ID.keys())-1)]
        return self.__graphElements_ID[ID]

    #Given a localId (w.r.t this graph), returns a reference
    #to the corresponding graph element. For instance:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # root.getElement("ER") ==> ok
    # root.getElement("ER.Entity") ==> ok
    # root.getElement("Entity") ==> NOT OK!
    # root.getElement("ER").getElement("Entity") ==> ok
    # IN: str, localId with respect to this graph
    # RETURN: GraphElement
    # Order: O(depth), or O(length(elem_ID))

    def getElement(self, elem_ID):
        from string import split
        fullPath=split(elem_ID, ".")
        if len(fullPath)==1:
            if not self.hasLocalId(elem_ID):
                raise InvalidIdException(elem_ID)
            return self.__graphElements_ID[elem_ID]
        else:
            curLevel=self
            for elem in fullPath:
                if isinstance(curLevel,Graph):
                    curLevel=curLevel.getElement(elem)
                else:
                    raise InvalidPathException(elem_ID,
                    "A node was found in the path, couldn't go deeper")
            return curLevel

    #Given a globalId, returns a graph element. For example:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # root.getElementGlobal("root.ER.Entity") ==> ok
    # root.getElementGlobal("ER.Entity") ==> NOT OK!
    # root.getElementGlobal("root.ER").getElementGlobal("root.ER.Entity") ==> ok
    # IN: str, globalId
    # RETURN: GraphElement
    # Order: O(depth)

    def getElementGlobal(self, globalId):
        from string import split
        fullPath=split(globalId, ".")
        fullPath=fullPath[1:]
        curLevel=self.getRoot()
        for elem in fullPath:
            if isinstance(curLevel,Graph):
                curLevel=curLevel.getElement(elem)
            else:
                raise InvalidPathException(elem_ID,
                "A node was found in the path, couldn't go deeper")
        return curLevel

    #Returns a collection of all the elements located in this graph
    # RETURN Collection (GraphElement)
    # Order: O(1)


    def getElements(self):
        return GraphElemCollection(self.__graphElements_ID,
                                   self.__graphElements)

    # Adds a GraphElement to this graph or to a specified deeper
    # graph. For instance:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # root.addElement(Graph("AA"), "ER") <-- adding the Graph "AA" in the graph "ER"
    # root.addElement(Node("BB"), "ER.AA") <-- adding the Node "BB" in the graph "AA"
    # root.addElement(Node("LA"), "ER.Entity") <-- NOT VALID, cannot add an
    #                                              an element in a Node!
    # IN_A: reference to a GraphElement
    # [IN_B]: potential localId that leads to a Graph in which "elem" will be added
    # RETURN: None
    # Order:  if elem is a Node O(length(path))
    #         if elem is a Graph O(num of edges+elements in elem + length(path))

    def addElement(self, elem, path=""):

        if path!="":
            parentGraph=self
            from string import split
            path=split(path, ".")
            for subgraph in path:
                if isinstance(parentGraph,Graph):
                    parentGraph=parentGraph.getElement(subgraph)
                else:
                    raise InvalidPathException(elem_ID,
                    "A node was found in the path, couldn't go deeper")

            if parentGraph!=self:
                parentGraph.addElement(elem)
                return

        ID = elem.getLocalId()

        if self.hasLocalId(ID):
            raise NonUniqueIdException(ID)

        if elem.getParent()==None:
            #If parent==None, it is either a node with no parent
            #or a root graph.  If it is a graph, we merge its
            #connections with ours (they will be unique, since
            #we verified that ID is unique). We set the parent
            #to "self" afterwards.

            connsTuple=None

            if isinstance(elem,Graph):
                connsTuple=elem.getConnTuples()

            elem.setGlobalId(self.getGlobalId()+"."+str(elem.getLocalId()))
            elem.setDepth(self.getDepth()+1)
            elem.setParent(self)
            elem.setRoot(self.getRoot())
            elem.createConnInfo()

            self.setAddElement(elem)

            if connsTuple != None:
                for conn in connsTuple:
                    elem.getRoot().connect(conn[0].getGlobalId(1),
                                           conn[1].getGlobalId(1))
        else:
            raise InvalidElementException(elem, "Cannot add an element "+\
                                          "from another graph, remove it first "+\
                                          "or use the popElement method")
            return

    # NOT TESTED... getLocalConn is NOT ok. should create a function
    # getInternalConn


    def addElements(self, elem, path=""):

        if not isinstance(elem,Graph):
            raise InvalidElementException(elem, "Node has no elements,"+\
                                          "Need a Graph, not a Node")

        if path!="":
            parentGraph=self
            from string import split
            path=split(path, ".")
            for subgraph in path:
                if isinstance(parentGraph,Graph):
                    parentGraph=parentGraph.getElement(subgraph)
                else:
                    raise InvalidPathException(elem_ID,
                    "A node was found in the path, couldn't go deeper")

            if parentGraph!=self:
                parentGraph.addElements(elem)
                return

        if elem.getParent()==None:
            connsTuple=elem.getConnTuples()
            elemIter=elem.getElements().iterate()

            while not elemIter.isDone():
                el=elemIter.currentElement()
                elemIter.next()
                if self.hasLocalId(el.getLocalId()):
                    raise NonUniqueIdException(ID)

            elemIter.first()
            while not elemIter.isDone():
                el=elemIter.currentElement()
                elemIter.next()
                elem.updateRemoveElem(el.getLocalId())
                el.setGlobalId(self.getGlobalId()+"."+str(el.getLocalId()))
                el.setDepth(self.getDepth()+1)
                el.setParent(self)
                el.setRoot(self.getRoot())
                el.createConnInfo()
                self.setAddElement(el)

            for conn in connsTuple:
                self.getRoot().connect(conn[0].getGlobalId(1),
                                       conn[1].getGlobalId(1))
        else:
            raise InvalidElementException(elem, "Cannot add an element "+\
                                          "from another graph, remove it first "+\
                                          "or use the moveElements method")

    # Given a localId, pop the corresponding element out of the Graph.
    # Create a new context for the element where it is becoming itself
    # a root for a new independent graph.  The element is removed from
    # this graph AND it is returned in a usable form.  For instance, it
    # is possible to:
    # root=Graph("root")
    # root.addElement(Graph("ER"))  <-- adding the graph "ER" in "root"
    # root.addElement(Node("Entity"), "ER")  <-- adding the Node "Entity" in "ER"
    # root.addElement(root.popElement("ER"))  <-- popping and adding back...
    # Some connections that are not meaningful to the new context of the popped
    # element will be removed.  i.e. Only internal connections of will be kept,
    # connections where the common parent is located within the popped element.
    # IN: str, localId
    # RETURN: GraphElement
    # Order O(edges+elements)

    def popElement(self, ID):

        from string import split
        fullPath=split(ID, ".")

        ID=fullPath[-1]

        curGraph=self
        for i in range(len(fullPath)-1):
            if isinstance(curGraph,Node):
                raise InvalidPathException(curGraph.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            curGraph=curGraph.getElement(fullPath[i])

        if not curGraph.hasLocalId(ID):
            raise InvalidIdException(ID)
        else:
            return curGraph.updatePopElem(ID)

    # Given a localId, remove the corresponding GraphElement out of this
    # graph
    # IN: str, localId
    # Order(element+edges)

    def removeElement(self, ID):

        from string import split
        fullPath=split(ID, ".")

        ID=fullPath[-1]

        curGraph=self
        for i in range(len(fullPath)-1):
            if isinstance(curGraph,Node):
                raise InvalidPathException(curGraph.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            curGraph=curGraph.getElement(fullPath[i])

        if not curGraph.hasLocalId(ID):
            raise InvalidIdException(ID)
        else:
            curGraph.updateRemoveElem(ID)

    # Given a localId, returns the outConnections of the corresponding
    # element as a Collection.
    # IN: str, localId
    # RETURN: Collection (GraphElement)
    # Order: O(depth)

    def getElemOutC(self, localId):

        from string import split
        fullPath=split(localId, ".")

        curGraph=self
        for i in range(len(fullPath)-1):
            if isinstance(curGraph,Node):
                raise InvalidPathException(curGraph.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            curGraph=curGraph.getElement(fullPath[i])

        return curGraph.outC()

    # Given a localId, returns the inConnections of the corresponding
    # element as a Collection.
    # IN: str, localId
    # RETURN: Collection (GraphElement)
    # Order: O(depth)

    def getElemInC(self, localId):

        from string import split
        fullPath=split(localId, ".")

        curGraph=self
        for i in range(len(fullPath)-1):
            if isinstance(curGraph,Node):
                raise InvalidPathException(curGraph.getLocalId(),
                      "A node was found in the path, couldn't go deeper")
            curGraph=curGraph.getElement(fullPath[i])

        return curGraph.inC()

    # Returns the out connections of this element as a Collection
    # RETURN: Collection (GraphElement)
    # Order: O(1)

    def outC(self):
        if self.__root==None:
            return GraphConnCollection({})

        conns=self.getRoot().getConnInfo()
        globalId=self.getGlobalId()

        return GraphConnCollection(conns[globalId][1])

    # Returns the in connections of this element as a Collection
    # RETURN: Collection (GraphElement)
    # Order: O(1)

    def inC(self):
        if self.__root==None:
            return GraphConnCollection({})

        conns=self.getRoot().getConnInfo()
        globalId=self.getGlobalId()

        return GraphConnCollection(conns[globalId][0])

    #This function is awfully slow O(e^2n^2), need to be optimized at some point
    def equals(self, elem):
        if not isinstance(elem,Graph):
            return 0

        if self.getLocalId() != elem.getLocalId():
            return 0

        if not (self.getType()==self and elem.getType()==elem):
            if not self.getType().equals(elem.getType()):
                return 0

        if self.getElements().size() != elem.getElements().size():
            return 0

        selfElems=self.getElements().iterate()
        elemElems=elem.getElements().iterate()

        while not selfElems.isDone():
            curSelf=selfElems.currentElement()
            while not elemElems.isDone():
                curElem=elemElems.currentElement()
                if curElem.equals(curSelf):
                    break
                elemElems.next()
            if elemElems.isDone():
                return 0
            selfElems.next()
            elemElems.first()

        selfConns=self.getLocalConn()
        elemConns=elem.getLocalConn()

        if len(selfConns) != len(elemConns):
            return 0

        for selfConn in selfConns:
            check=0
            for elemConn in elemConns:
                if selfConn[0].equals(elemConn[0]):
                    if selfConn[1].equals(elemConn[1]):
                        if selfConn[2]==elemConn[2]:
                            if selfConn[3]==elemConn[3]:
                                check=1
                                break
            if not check:
                return 0

        return 1

    def clone(self):

        new = Graph(self.getLocalId(),
                    parent=None,
                    type=self.getTypeLocalId())

        selfElems=self.getElements().iterate()
        while not selfElems.isDone():
            curSelf=selfElems.currentElement()
            new.addElement(curSelf.clone())
            selfElems.next()

        conns=self.getLocalConn()
        for conn in conns:
            new.connect(conn[0].getGlobalId(self.getDepth()+1),
                        conn[1].getGlobalId(self.getDepth()+1),
                        conn[2],
                        conn[3])
        return new


    def __repr__(self):
        return str(self.__globalId)

    def __str__(self):

        answer = str(self.__globalId) + " :: " + str(self.getTypeLocalId())
        return answer

    def __del__(self):
        #print "del graph", self
        del self.__id
        del self.__type
        del self.__graphElements
        del self.__graphElements_ID
        del self.__parent
        del self.__depth
        del self.__globalId
        if self.__root==None:
            del self.__elemConn
        del self.__root

    # enter action perfomed by a visitor on this element
    def enter(self, visitor):
        visitor.enterGraph(self)

    # exit action perfomed by a visitor on this element
    def exit(self, visitor):
        visitor.exitGraph(self)
